/** @file wxw_screen.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2014.03.21

	@brief [ wxw screen ]

	@note
	This code is based on win32_screen.cpp of the Common Source Code Project.
	Author : Takeda.Toshiya
*/

#include <wx/wx.h>
#include <wx/display.h>
#include <wx/image.h>
#include <wx/rawbmp.h>
#include "wxw_emu.h"
#include "../../vm/vm.h"
#ifdef USE_MESSAGE_BOARD
#include "../../msgboard.h"
#endif
#include "../../csurface.h"
#include "../../video/rec_video.h"
#include "wxw_main.h"
#include "../../vm/vm.h"
#include "../../gui/gui.h"
#include "../../config.h"
#include "../../utils.h"

//static wxVideoMode screen_mode_wx[SCREEN_MODE_MAX];

static int window_pos_x = -1;
static int window_pos_y = -1;

void EMU_OSD::EMU_SCREEN()
{
	scrBmp = new CSurface();
	screen_flags = 0;

#ifdef USE_OPENGL
	mix_texture_name = 0;
	src_tex_l = src_tex_t = 0.0;
	src_tex_r = src_tex_b = 1.0;
#endif
}

///
/// initialize screen
///
void EMU_OSD::initialize_screen()
{
	EMU::initialize_screen();

#ifdef USE_OPENGL
//	initialize_opengl();
#endif

#ifdef USE_MESSAGE_BOARD
	msgboard = new MsgBoard(this);
	if (msgboard) {
		msgboard->InitScreen(pixel_format, screen_size.w, screen_size.h);
		msgboard->SetMessagePos(4 + reMix.x, -4 - reMix.y, 2);
		msgboard->SetInfoPos(-4 - reMix.x, 4 + reMix.y, 1);
		msgboard->SetVisible(FLG_SHOWMSGBOARD ? true : false);
	}
#endif
	if (gui) {
		gui->CreateLedBox(res_path, pixel_format);
	}
}

///
/// release screen
///
void EMU_OSD::release_screen()
{
#ifdef USE_OPENGL
	release_opengl();
#endif
	if (gui) {
		gui->ReleaseLedBox();
	}

	delete scrBmp;
	scrBmp = NULL;

	EMU::release_screen();
}

///
/// change pixel format
///
void EMU_OSD::change_pixel_format()
{
	if (pixel_format != NULL) {
		// change pixel format
#if defined(__WXMSW__)
#  ifdef USE_OPENGL
		if (config.use_opengl) {
			pixel_format->PresetRGBA();
		} else
#  endif
		{
			pixel_format->PresetBGRA();
		}
#elif defined(__WXOSX__)
#  ifdef USE_OPENGL
		if (config.use_opengl) {
			pixel_format->PresetARGB();
		} else
#  endif
		{
			pixel_format->PresetARGB();
		}
#elif defined(__WXGTK__)
#  ifdef USE_OPENGL
		if (config.use_opengl) {
			pixel_format->PresetRGBA();
		} else
#  endif
		{
			pixel_format->PresetBGRA();
		}
#else
		pixel_format->PresetRGBA();
#endif
	}
}

///
/// create / recreate window
///
bool EMU_OSD::create_screen(int disp_no, int x, int y, int width, int height, uint32_t flags)
{
	screen_flags = flags;

	change_pixel_format();

#ifdef _DEBUG_LOG
	logging->out_debugf(_T("VideoInfo Rmask:%x, Gmask:%x, Bmask:%x, Amask:%x"),pixel_format->Rmask,pixel_format->Gmask,pixel_format->Bmask,pixel_format->Amask);
	logging->out_debugf(_T("VideoInfo Rshift:%d, Gshift:%d, Bshift:%d, Ashift:%d"),pixel_format->Rshift,pixel_format->Gshift,pixel_format->Bshift,pixel_format->Ashift);
#endif

	return true;
}

///
/// create / recreate offline surface
///
bool EMU_OSD::create_offlinesurface()
{
	if (!pixel_format) return false;

	if (sufOrigin && !sufOrigin->IsEnable()) {
		if (!sufOrigin->Create(screen_size.w, screen_size.h, *pixel_format)) {
			logging->out_log(LOG_ERROR, _T("EMU_OSD::create_offlinesurface sufOrigin failed."));
			return false;
		}
	}
#ifdef USE_SCREEN_ROTATE
	if (sufRotate) {
		sufRotate->Release();
		if (!sufRotate->Create(source_size.w, source_size.h, *pixel_format)) {
			logging->out_log(LOG_ERROR, _T("EMU_OSD::create_offlinesurface sufRotate failed."));
			return false;
		}
	}
#endif
	if (sufMixed) {
		sufMixed->Release();
		if (!sufMixed->Create(mixed_max_size.w, mixed_max_size.h, *pixel_format)) {
			logging->out_log(LOG_ERROR, _T("EMU_OSD::create_offlinesurface sufMixed failed."));
			return false;
		}
	}
#ifdef USE_SMOOTH_STRETCH
	if (sufStretch1 && sufStretch2) {
		sufStretch1->Release();
		sufStretch2->Release();
		stretch_screen = false;
		if(stretch_power.w != 1 || stretch_power.h != 1) {
			sufStretch1->Create(source_size.w * stretch_power.w, source_size.h * stretch_power.h);
			SetStretchBltMode(sufStretch1->GetDC(), COLORONCOLOR);

			sufStretch2->Create(stretched_size.w, stretched_size.h);
			SetStretchBltMode(sufStretch2->GetDC(), COLORONCOLOR);

			stretch_screen = true;
		}
	}
#endif

	disable_screen &= ~DISABLE_SURFACE;

	return true;
}

///
/// setting window or fullscreen size
///
/// @param [in] width : new width or -1 set current width
/// @param [in] height : new height or -1 set current height
/// @param [in] power : magnify x 10
/// @param [in] window_mode : true:window / false:fullscreen
void EMU_OSD::set_display_size(int width, int height, int power, bool window_mode)
{
	bool display_size_changed = false;
	bool stretch_changed = false;

	if(width != -1 && (display_size.w != width || display_size.h != height)) {
		display_size.w = width;
		display_size.h = height;
		display_size_changed = stretch_changed = true;
	}

#ifdef USE_SCREEN_ROTATE
	VmRectWH prev_source_size = source_size;

	if(config.monitor_type & 1) {
		stretch_changed |= (source_size.w != screen_size.h);
		stretch_changed |= (source_size.h != screen_size.w);
		stretch_changed |= (source_aspect_size.w != screen_aspect_size.h);
		stretch_changed |= (source_aspect_size.h != screen_aspect_size.w);

		source_size.w = screen_size.h;
		source_size.h = screen_size.w;
		source_aspect_size.w = screen_aspect_size.h;
		source_aspect_size.h = screen_aspect_size.w;
	} else
#endif
	{
		stretch_changed |= (source_size.w != screen_size.w);
		stretch_changed |= (source_size.h != screen_size.h);
		stretch_changed |= (source_aspect_size.w != screen_aspect_size.w);
		stretch_changed |= (source_aspect_size.h != screen_aspect_size.h);

		source_size = screen_size;
		source_aspect_size = screen_aspect_size;
	}

	// fullscreen and stretch screen
	if(config.stretch_screen && !window_mode && display_size.w >= source_size.w && display_size.h >= source_size.h) {
		if (config.stretch_screen == 1) {
			// fit to full screen
			mixed_size = source_size;
			if (mixed_ratio.w < mixed_ratio.h) {
				mixed_size.h = mixed_size.h * mixed_ratio.w / mixed_ratio.h;
			} else {
				mixed_size.w = mixed_size.w * mixed_ratio.h / mixed_ratio.w;
			}
			mixed_size.x = (source_size.w - mixed_size.w) / 2;
			mixed_size.y = (source_size.h - mixed_size.h) / 2;
			mixed_size.y = adjust_y_position(mixed_size.h, mixed_size.y);

			stretched_size.w = (display_size.h * source_aspect_size.w) / source_aspect_size.h;
			stretched_size.h = display_size.h;
			stretched_dest_real.x = - mixed_size.x * display_size.h / source_aspect_size.h;
			stretched_dest_real.y = - mixed_size.y * display_size.h / source_aspect_size.h;
			if(stretched_size.w > display_size.w) {
				stretched_size.w = display_size.w;
				stretched_size.h = (display_size.w * source_aspect_size.h) / source_aspect_size.w;
				stretched_dest_real.x = - mixed_size.x * display_size.w / source_aspect_size.w;
				stretched_dest_real.y = - mixed_size.y * display_size.w / source_aspect_size.w;
			}
			if (mixed_ratio.w < mixed_ratio.h) {
				stretched_dest_real.y = stretched_dest_real.y * mixed_ratio.h / mixed_ratio.w;
			} else {
				stretched_dest_real.x = stretched_dest_real.x * mixed_ratio.w / mixed_ratio.h;
			}
			stretched_size.x = (display_size.w - stretched_size.w) / 2;
			stretched_size.y = (display_size.h - stretched_size.h) / 2;
			stretched_dest_real.x += stretched_size.x;
			stretched_dest_real.y += stretched_size.y;

		} else {
			// fit text area to full screen (cut off padding area)
			VmSize min_size;
			SIZE_IN(min_size, LIMIT_MIN_WINDOW_WIDTH, LIMIT_MIN_WINDOW_HEIGHT);
#ifdef USE_SCREEN_ROTATE
			if(config.monitor_type & 1) {
				SWAP(int, min_size.w, min_size.h);
			}
#endif
			VmSize mixed_rsize, mixed_rrsize;
			mixed_size = source_size;
			SIZE_IN(mixed_rsize, mixed_size.w, mixed_size.h);
			mixed_rrsize = mixed_rsize;

			VmSize min_rsize, min_rrsize;
			min_rsize = min_size;
			min_rrsize = min_size;
			if (mixed_ratio.w < mixed_ratio.h) {
				min_rsize.h = min_rsize.h * mixed_ratio.h / mixed_ratio.w;
				min_rrsize.h = min_rrsize.h * mixed_ratio.w / mixed_ratio.h;
				mixed_rsize.h = mixed_rsize.h * mixed_ratio.h / mixed_ratio.w;
				mixed_rrsize.h = mixed_rrsize.h * mixed_ratio.w / mixed_ratio.h;
			} else {
				min_rsize.w = min_rsize.w * mixed_ratio.w / mixed_ratio.h;
				min_rrsize.w = min_rrsize.w * mixed_ratio.h / mixed_ratio.w;
				mixed_rsize.w = mixed_rsize.w * mixed_ratio.w / mixed_ratio.h;
				mixed_rrsize.w = mixed_rrsize.w * mixed_ratio.h / mixed_ratio.w;
			}

			double magx = (double)display_size.w / min_rsize.w;
			double magy = (double)display_size.h / min_rsize.h;

			bool mag_based_w = (magx < magy);

			stretched_dest_real.x = 0;
			stretched_dest_real.y = 0;
			if(mag_based_w) {
				// magnify = display_size.w / min_rsize.w

				mixed_size.x = (mixed_size.w - min_size.w) / 2;
				mixed_size.w = min_size.w;
				stretched_size.x = 0;
				stretched_size.w = display_size.w;
				stretched_dest_real.x = (display_size.w - (mixed_rsize.w * display_size.w / min_rsize.w)) / 2;

				if ((mixed_size.h * min_rsize.h * display_size.w / min_rsize.w / min_size.h) >= display_size.h) {
					int mh = display_size.h * min_rrsize.h * min_rsize.w / display_size.w / min_size.h;
					mixed_size.y = (mixed_size.h - mh) / 2;
					mixed_size.h = mh;
					stretched_size.y = 0;
					stretched_size.h = display_size.h;
					stretched_dest_real.y = (display_size.h - (mixed_size.h * display_size.w / min_rsize.w)) / 2;
				} else {
					int mh = mixed_rrsize.h;
					int sh = mixed_size.h * display_size.w / min_rsize.w;
					mixed_size.y = (mixed_size.h - mh) / 2;
					mixed_size.h = mh;
					stretched_size.y = (display_size.h - sh) / 2;
					stretched_size.h = sh;
					stretched_dest_real.y = stretched_size.y;
				}

			} else {
				// magnify = display_size.h / min_rsize.h

				mixed_size.y = (mixed_size.h - min_size.h) / 2;
				mixed_size.h = min_size.h;
				stretched_size.y = 0;
				stretched_size.h = display_size.h;
				stretched_dest_real.y = (display_size.h - (mixed_rsize.h * display_size.h / min_rsize.h)) / 2;

				if ((mixed_size.w * min_rsize.w * display_size.h / min_rsize.h / min_size.w) >= display_size.w) {
					int mw = display_size.w * min_rrsize.w * min_rsize.h / display_size.h / min_size.w;
					mixed_size.x = (mixed_size.w - mw) / 2;
					mixed_size.w = mw;
					stretched_size.x = 0;
					stretched_size.w = display_size.w;
					stretched_dest_real.x = (display_size.w - (mixed_rsize.w * display_size.h / min_rsize.h)) / 2;
				} else {
					int mw = mixed_rrsize.w;
					int sw = mixed_size.w * display_size.h / min_rsize.h;
					mixed_size.x = (mixed_size.w - mw) / 2;
					mixed_size.w = mw;
					stretched_size.x = (display_size.w - sw) / 2;
					stretched_size.w = sw;
					stretched_dest_real.x = stretched_size.x;
				}

			}
		}
	}
	// window or non-streach mode
	else {
		for(int n = 0; n <= 1; n++) {
			if (n == 0) {
				mixed_size.w = display_size.w * 10 / power;
				mixed_size.h = display_size.h * 10 / power;
			} else {
				mixed_size = source_size;
			}
			if (mixed_ratio.w < mixed_ratio.h) {
				mixed_size.h = mixed_size.h * mixed_ratio.w / mixed_ratio.h;
			} else {
				mixed_size.w = mixed_size.w * mixed_ratio.h / mixed_ratio.w;
			}
			mixed_size.x = (source_size.w - mixed_size.w) / 2;
			mixed_size.y = (source_size.h - mixed_size.h) / 2;
			if (mixed_size.x >= 0 && mixed_size.y >= 0) {
				break;
			}
		}
		mixed_size.y = adjust_y_position(mixed_size.h, mixed_size.y);

		stretched_size.w = source_aspect_size.w * power / 10;
		stretched_size.h = source_aspect_size.h * power / 10;
		stretched_size.x = (display_size.w - stretched_size.w) / 2;
		stretched_size.y = (display_size.h - stretched_size.h) / 2;
		stretched_dest_real.x = - mixed_size.x * power / 10;
		stretched_dest_real.y = - mixed_size.y * power / 10;
		if (mixed_ratio.w < mixed_ratio.h) {
			stretched_dest_real.y = stretched_dest_real.y * mixed_ratio.h / mixed_ratio.w;
		} else {
			stretched_dest_real.x = stretched_dest_real.x * mixed_ratio.w / mixed_ratio.h;
		}
		if (stretched_size.x < 0) {
			stretched_size.x = 0;
			stretched_size.w = display_size.w;
		}
		if (stretched_size.y < 0) {
			stretched_size.y = 0;
			stretched_size.h = display_size.h;
		}
		stretched_dest_real.x += stretched_size.x;
		stretched_dest_real.y += stretched_size.y;
	}

	// sufMixed size normalize 2^n for opengl
	mixed_max_size.w = 0;
	mixed_max_size.h = 0;
	// 65536,32768,16384,8192,4096,2048,1024,512,256,128,64,32,16
	for(int i=16; i>=4; i--) {
		if (source_size.w <= (1 << i)) mixed_max_size.w = (1 << i);
		if (source_size.h <= (1 << i)) mixed_max_size.h = (1 << i);
	}

	reMix = mixed_size;
	reSuf = stretched_size;

#ifdef USE_OPENGL
	src_tex_l = (GLdouble)mixed_size.x / mixed_max_size.w;
	src_tex_t = (GLdouble)mixed_size.y / mixed_max_size.h;
	src_tex_r = (GLdouble)(mixed_size.x + mixed_size.w) / mixed_max_size.w;
	src_tex_b = (GLdouble)(mixed_size.y + mixed_size.h) / mixed_max_size.h;
	reTex = stretched_size;
#endif

	change_rec_video_size(config.screen_video_size);

	first_invalidate = true;
	screen_size_changed = false;
#ifdef _DEBUG_LOG
	logging->out_debugf(_T("set_display_size: w:%d h:%d power:%d %s"),width,height,power,window_mode ? _T("window") : _T("fullscreen"));
	logging->out_debugf(_T("         display: w:%d h:%d"),display_size.w, display_size.h);
	logging->out_debugf(_T("          screen: w:%d h:%d"), screen_size.w, screen_size.h);
	logging->out_debugf(_T("   screen aspect: w:%d h:%d"), screen_aspect_size.w, screen_aspect_size.h);
	logging->out_debugf(_T("          source: w:%d h:%d"), source_size.w, source_size.h);
	logging->out_debugf(_T("   source aspect: w:%d h:%d"), source_aspect_size.w, source_aspect_size.h);
	logging->out_debugf(_T("           mixed: w:%d h:%d"), mixed_size.w, mixed_size.h);
	logging->out_debugf(_T("         stretch: w:%d h:%d"), stretched_size.w, stretched_size.h);
	logging->out_debugf(_T("      mixed dest: x:%d y:%d"), mixed_size.x, mixed_size.y);
	logging->out_debugf(_T("    stretch dest: x:%d y:%d"), stretched_size.x, stretched_size.y);
	logging->out_debugf(_T(" stretch dest re: x:%d y:%d"), stretched_dest_real.x, stretched_dest_real.y);
#ifdef USE_OPENGL
	logging->out_debugf(_T("     src texture: l:%.5f t:%.5f r:%.5f b:%.5f"), src_tex_l, src_tex_t, src_tex_r, src_tex_b);
	logging->out_debugf(_T("         texture: x:%d y:%d w:%d h:%d"), reTex.x, reTex.y, reTex.w, reTex.h);
#endif
#endif
	if (window_mode) {
		stretched_size.x += display_margin.left;
		stretched_size.y += display_margin.top;
		reSuf.x += display_margin.left;
		reSuf.y += display_margin.top;
#ifdef _DEBUG_LOG
		logging->out_debugf(_T(" margin      suf: l:%d t:%d w:%d h:%d"),reSuf.x,reSuf.y,reSuf.w,reSuf.h);
#endif
#ifdef USE_OPENGL
		reTex.x += display_margin.left;
		reTex.y += display_margin.top;
#ifdef _DEBUG_LOG
		logging->out_debugf(_T("             tex: l:%d t:%d w:%d h:%d"),reTex.x,reTex.y,reTex.w,reTex.h);
#endif
#endif
	}

	lock_screen();

#ifdef USE_SCREEN_ROTATE
	if(config.monitor_type) {
		sufSource = sufRotate;
	} else
#endif
	{
		sufSource = sufOrigin;
	}

#ifdef USE_SCREEN_ROTATE
	if ((source_size.w != prev_source_size.w && source_size.h != prev_source_size.h) || stretch_changed) {
		create_offlinesurface();
	}
#endif

	// send display size to vm
	set_vm_display_size();

	if (gui) {
		gui->SetLedBoxPosition(window_mode, mixed_size.x, mixed_size.y, mixed_size.w, mixed_size.h, config.led_pos | (is_fullscreen() ? 0x10 : 0));
	}
#ifdef USE_MESSAGE_BOARD
	if (msgboard) {
		msgboard->SetSize(source_size.w, source_size.h);
		msgboard->SetMessagePos(4 + mixed_size.x,  - 4 - source_size.h + mixed_size.y + mixed_size.h, 2);
		msgboard->SetInfoPos(-4 - mixed_size.x, 4 + mixed_size.y, 1);
	}
#endif

	unlock_screen();
}

///
/// draw src screen from virtual machine
///
void EMU_OSD::draw_screen()
{
	// don't draw screen before new screen size is applied to buffers
	if(screen_size_changed) {
		return;
	}

	lock_screen();

	if (sufOrigin->Lock()) {
		if (!config.now_power_off) {
			vm->draw_screen();
		} else {
			fill_gray();
		}
		sufOrigin->Unlock();
	}

	// screen size was changed in vm->draw_screen()
	if(screen_size_changed) {
		unlock_screen();
		return;
	}

#ifdef USE_SCREEN_ROTATE
	// rotate screen
	// right turn
	// src and dst should be the same size
	if(config.monitor_type) {
		int rtype = (config.monitor_type & 3);
		VmSize ss, ds;
		SIZE_IN(ss, screen_size.w, screen_size.h);
		SIZE_IN(ds, source_size.w, source_size.h);
		{
			int nw0 = rotate_matrix[rtype].x[0] < 0 ? (ss.w - 1) : 0;
			int nh0 = rotate_matrix[rtype].y[0] < 0 ? (ss.h - 1) : 0;
			int nw1 = rotate_matrix[rtype].x[1] < 0 ? (ss.w - 1) : 0;
			int nh1 = rotate_matrix[rtype].y[1] < 0 ? (ss.h - 1) : 0;
			for(int sy = 0; sy < ss.h; sy++) {
				scrntype* src = sufOrigin->GetBuffer() + ss.w * sy;
				for(int sx = 0; sx < ss.w; sx++) {
					int dx = nw0 + rotate_matrix[rtype].x[0] * sx + nh0 + rotate_matrix[rtype].y[0] * sy; 
					int dy = nw1 + rotate_matrix[rtype].x[1] * sx + nh1 + rotate_matrix[rtype].y[1] * sy;
					scrntype* dst = sufRotate->GetBuffer() + ds.w * dy + dx;
					*dst = *src;
					src++;
				}
			}
		}
	}
#endif

	{
		if (FLG_SHOWLEDBOX && gui) {
			gui->DrawLedBox(sufSource);
		}
	}

	unlock_screen();
}

///
/// copy src screen to mix screen and overlap a message
///
/// @return false: cannot mix (no allocate mix surface)
bool EMU_OSD::mix_screen()
{
	if (disable_screen) return false;

	lock_screen();

	sufSource->Blit(*sufMixed);

	unlock_screen();

	if (gui) {
		gui->UpdateIndicator(update_led());
	}
#ifdef USE_MESSAGE_BOARD
	if (msgboard && FLG_SHOWMSGBOARD) {
		msgboard->Draw(sufMixed);
	}
#endif
	return true;
}

///
/// update GLCanvas
///
void EMU_OSD::update_screen_gl(MyGLCanvas *panel)
{
	gui->UpdateScreen();

	if (mix_screen()) {

#ifdef USE_OPENGL
		if (mix_texture_name > 0) {

			GLenum gl_errno;
#if defined(__WXMSW__)
			scrntype *buffer = sufMixed->GetBuffer();
			sufMixed->Flip();
//			buffer -= ((sufMixed->Height() - 1) * sufMixed->Width());
#elif defined(__WXOSX__)
			char *buffer = (char *)sufMixed->GetBuffer();
//			buffer -= (sufMixed->Width() * sufMixed->BytesPerPixel());
			buffer++;
#else
			scrntype *buffer = sufMixed->GetBuffer();
#endif
			// draw texture using screen pixel buffer
			glBindTexture(GL_TEXTURE_2D, mix_texture_name);

			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, sufMixed->Width(), sufMixed->Height(), 0, GL_RGBA,
					GL_UNSIGNED_BYTE, (GLvoid *)buffer);
			if ((gl_errno = glGetError()) != GL_NO_ERROR) {
				logging->out_logf(LOG_ERROR, _T("glTexImage2D: 0x%x"), gl_errno);
			}

			// fill black
			glBegin(GL_QUADS);
				glColor3ub(0, 0, 0);
				glVertex3i(0, 0, 0);
				glVertex3i(0, display_size.h, 0);
				glVertex3i(display_size.w, display_size.h, 0);
				glVertex3i(display_size.w, 0, 0);
			glEnd();

			glEnable(GL_TEXTURE_2D);
			glColor3ub(255, 255, 255);
			glBegin(GL_QUADS);
				// anticlockwise
				glTexCoord2d(src_tex_l, src_tex_t);
				glVertex3i(reTex.x, reTex.y, 0);
				glTexCoord2d(src_tex_l, src_tex_b);
				glVertex3i(reTex.x, reTex.y + reTex.h, 0);
				glTexCoord2d(src_tex_r, src_tex_b);
				glVertex3i(reTex.x + reTex.w, reTex.y + reTex.h, 0);
				glTexCoord2d(src_tex_r, src_tex_t);
				glVertex3i(reTex.x + reTex.w, reTex.y, 0);
			glEnd();
			glDisable(GL_TEXTURE_2D);

			glBindTexture(GL_TEXTURE_2D, 0);

			glFlush();

			panel->SwapBuffers();

			// invalidate window
			self_invalidate = true;
			skip_frame = false;
		}
#endif
	}
	gui->UpdatedScreen();
	return;
}

#ifdef __WXOSX__
//#define USE_BUFFERD_PAINT_DC
#endif
#ifdef USE_BUFFERD_PAINT_DC
#include <wx/dcbuffer.h>
#endif

void EMU_OSD::update_screen_pa(MyPanel *panel)
{
	gui->UpdateScreen();

	if (mix_screen()) {

#ifdef USE_BUFFERD_PAINT_DC
#ifdef __WXOSX__
		wxImage img(display_width, display_height);
		if (stretched_size.w == mixed_size.w && stretched_size.h == mixed_size.h) {
			sufMixed->Blit(reMix, img, reSuf);
		} else {
			sufMixed->StretchBlit(reMix, img, reSuf);
		}
		wxBitmap bmp(img);
#else
		wxBitmap bmp(display_width, display_height, 24);
		if (stretched_size.w == mixed_size.w && stretched_size.h == mixed_size.h) {
			sufMixed->Blit(reMix, bmp, reSuf);
		} else {
			sufMixed->StretchBlit(reMix, bmp, reSuf);
		}
#endif
#else
		if (stretched_size.w == mixed_size.w && stretched_size.h == mixed_size.h) {
			sufMixed->Blit(reMix, *scrBmp, reSuf);
		} else {
			sufMixed->StretchBlit(reMix, *scrBmp, reSuf);
		}
#endif

#ifdef USE_BUFFERD_PAINT_DC 
		// paint the screen
		wxBufferedPaintDC dc(panel, bmp);
//		if (!dc.IsOk()) {
//			logging->out_log(LOG_ERROR,_T("update_screen: bufferd paint dc failed."));
//		}
#else
		wxPaintDC dc(panel);
		dc.Blit(0, 0, display_size.w, display_size.h, scrBmp->GetDC(), 0, 0);
#endif

		// invalidate window
		self_invalidate = true;
		skip_frame = false;
	}
	gui->UpdatedScreen();
	return;
}

///
/// post request screen updating to draw it on main thread
///
//void EMU_OSD::need_update_screen()
//{
//	// nothing to do
//}

///
/// pointer on source screen
///
/// @return pointer on source screen
scrntype* EMU_OSD::screen_buffer(int y)
{
	return sufOrigin->GetBuffer(y);
}

///
/// offset on source screen
///
/// @return offset on source screen
int EMU_OSD::screen_buffer_offset()
{
#if defined(__WXMSW__)
	return -screen_size.w;
#else
	return screen_size.w;
#endif
}

///
/// capture current screen and save to a file
///
void EMU_OSD::capture_screen()
{
	int size = config.screen_video_size;
	rec_video->Capture(CAPTURE_SCREEN_TYPE, rec_video_stretched_size, sufSource, rec_video_size[size]);
}

///
/// start recording video
///
//bool EMU_OSD::start_rec_video(int type, int fps_no, bool show_dialog)
//{
//#ifdef USE_REC_VIDEO
//	int size = config.screen_video_size;
//	return rec_video->Start(type, fps_no, rec_video_size[size], sufSource, show_dialog);
//#else
//	return false;
//#endif
//}

///
/// record video
///
void EMU_OSD::record_rec_video()
{
#ifdef USE_REC_VIDEO
	if (rec_video->IsRecordFrame()) {
		int size = config.screen_video_size;
		rec_video->Record(rec_video_stretched_size, sufSource, rec_video_size[size]);
	}
#endif
}

//void EMU_OSD::enum_screen_mode(uint32_t flags)
//{
//	screen_mode.Enum(desktop_size.w, desktop_size.h, flags);
//}

/// change window size / switch over fullscreen and window
/// @param[in] mode 0 - 7: window size  8 -:  fullscreen size  -1: switch over  -2: shift window mode
void EMU_OSD::change_screen_mode(int mode)
{
//	logging->out_debugf(_T("change_screen_mode: mode:%d cwmode:%d pwmode:%d w:%d h:%d"),mode,config.window_mode,prev_window_mode,desktop_width,desktop_height);
//	if (mode == config.window_mode) return;

	if (mode == -1) {
		// switch over fullscreen and window
		if (now_fullscreen) {
			// go window mode
			mode = prev_window_mode;
		}
	} else if (mode == -2) {
		// shift window mode 
		if (now_fullscreen) {
			// no change
			return;
		} else {
			mode = ((config.window_mode + 1) % window_mode.Count());
		}
	}
	if (!now_fullscreen) {
		prev_window_mode = config.window_mode;
	}
//	logging->out_debugf(_T("change_screen_mode: mode:%d cwmode:%d pwmode:%d w:%d h:%d"),mode,config.window_mode,prev_window_mode,desktop_width,desktop_height);
	set_window(mode, desktop_size.w, desktop_size.h);
	if (!create_screen(config.disp_device_no, 0, 0, config.screen_width, config.screen_height, screen_flags)) {
		exit(1);
	}

}

/// setting window or fullscreen size to display
/// @param [in] mode 0 .. 7 window mode / 8 .. 23 fullscreen mode / -1 want to go fullscreen, but unknown mode
/// @param [in] cur_width  current desktop width if mode is -1
/// @param [in] cur_height current desktop height if mode is -1
void EMU_OSD::set_window(int mode, int cur_width, int cur_height)
{
	logging->out_debugf(_T("set_window: mode:%d"), mode);
	if(mode >= 0 && mode < 8) {
		// go window
		if (mode >= window_mode.Count()) mode = 0;
		const window_mode_t *wm = window_mode.Get(mode);
		int width = wm->width;
		int height = wm->height;
#ifdef USE_SCREEN_ROTATE
		if (config.monitor_type & 1) {
			int v = width;
			width = height;
			height = v;
		}
#endif

		config.window_mode = mode;
		config.disp_device_no = 0;
		config.screen_width = width;
		config.screen_height = height;
		window_mode_power = wm->power;

		now_fullscreen = false;

		if (gui) {
			MyFrame *frame = gui->GetMyFrame();
//			int num = mode;
			if (window_pos_x < 0 || window_pos_y < 0) {
				// center
				frame->GetPosition(&window_pos_x, &window_pos_y);
//				window_pos_x = (desktop_width - width) / 2;
//				window_pos_y = (desktop_height - height) / 2;
			}
			if (frame->IsFullScreen()) {
				// go back to default mode
				wxDisplay disp(config.disp_device_no);
				disp.ChangeMode();
//				wxRect re = disp.GetGeometry();
				frame->ShowFullScreen(false);
				frame->SetWindowStyleFlag(wxDEFAULT_FRAME_STYLE);
				frame->SetSize(window_pos_x, window_pos_y, wxDefaultCoord, wxDefaultCoord, wxSIZE_USE_EXISTING);
			}
			frame->SetClientSize(wm->width, wm->height);

			if (scrBmp != NULL) {
//				delete scrDC;
				delete scrBmp;
			}
//			scrBmp = new wxBitmap(wm->width, wm->height, 24);
//			scrDC = new wxMemoryDC(*scrBmp);
			scrBmp = new CSurface(wm->width, wm->height, CPixelFormat::RGBA32);

			gui->ScreenModeChanged(now_fullscreen);
		}
		// set screen size to emu class
		set_display_size(width, height, window_mode_power, true);
	}
	else if(!now_fullscreen) {
		// go fullscreen
		if (mode >= screen_mode.CountMode() + 8) {
			mode = -1;
		}
		if(mode == -1) {
			// matching width and height
			int find = screen_mode.FindMode(0, cur_width, cur_height);
			if (find >= 0) {
				mode = find + 8;
			} else {
				mode = 8;
			}
		}
		const screen_mode_t *sm = screen_mode.GetMode(mode - 8);
		int disp_no = sm ? sm->disp_no : 0;
		int width = sm ? sm->width : cur_width;
		int height = sm ? sm->height : cur_height;

		config.window_mode = mode;
		config.disp_device_no = disp_no;
		config.screen_width = width;
		config.screen_height = height;

		now_fullscreen = true;

		if (gui) {
			MyFrame *frame = gui->GetMyFrame();
			frame->GetPosition(&window_pos_x, &window_pos_y);
//			int num = mode - WINDOW_MODE_MAX;
			wxDisplay disp(disp_no);
			wxRect re = disp.GetGeometry();
			wxVideoMode mode = disp.GetCurrentMode();
			mode.w = width;
			mode.h = height;
			disp.ChangeMode(mode);
			frame->SetWindowStyleFlag(0);
			frame->SetSize(re.x, re.y, wxDefaultCoord, wxDefaultCoord, wxSIZE_USE_EXISTING);
			frame->SetClientSize(width, height);
			frame->ShowFullScreen(true, wxFULLSCREEN_ALL);

			if (scrBmp != NULL) {
//				delete scrDC;
				delete scrBmp;
			}
//			scrBmp = new wxBitmap(width, height, 24);
//			scrDC = new wxMemoryDC(*scrBmp);
			scrBmp = new CSurface(width, height, CPixelFormat::RGBA32);

			gui->ScreenModeChanged(now_fullscreen);
		}

		// set screen size to emu class
		set_display_size(width, height, 10, false);
	}
}

// ----------
#ifdef USE_OPENGL

void EMU_OSD::initialize_opengl()
{
	if (!use_opengl) {
		return;
	}
	create_texture();
}

void EMU_OSD::create_texture()
{
	GLenum gl_errno;

	if (mix_texture_name != 0) return;

	// create
#if defined(_RGB555) || defined(_RGB565)
	glPixelStorei(GL_UNPACK_ALIGNMENT, 2);
#elif defined(_RGB888)
	glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
#else
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
#endif

	glGenTextures(1, &mix_texture_name);
	if ((gl_errno = glGetError()) != GL_NO_ERROR) {
		logging->out_logf(LOG_ERROR, _T("glGenTextures: %d"), gl_errno);
	} else {
		logging->out_debugf(_T("glGenTextures: name:%d"), mix_texture_name);
	}
	glBindTexture(GL_TEXTURE_2D, mix_texture_name);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST + config.gl_filter_type);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST + config.gl_filter_type);
	glTexParameteri(GL_TEXTURE_2D , GL_TEXTURE_WRAP_S, GL_CLAMP);	// no wrap texture
	glTexParameteri(GL_TEXTURE_2D , GL_TEXTURE_WRAP_T, GL_CLAMP);
//	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, SCREEN_WIDTH, SCREEN_HEIGHT, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
//	if ((gl_errno = glGetError()) != GL_NO_ERROR) {
//		logging->out_logf(LOG_ERROR, "glTexImage2D: %d", gl_errno);
//	}
	glBindTexture(GL_TEXTURE_2D, 0);	// unbound

	// fill black
	glBegin(GL_QUADS);
	glColor3ub(0, 0, 0);
	glVertex3i(0, 0, 0);
	glVertex3i(0, display_size.h, 0);
	glVertex3i(display_size.w, display_size.h, 0);
	glVertex3i(display_size.w, 0, 0);
	glEnd();
}

void EMU_OSD::change_opengl_attr()
{
	if (!use_opengl) {
		return;
	}

	if (mix_texture_name != 0 && glIsTexture(mix_texture_name) == GL_TRUE) {
		glBindTexture(GL_TEXTURE_2D, mix_texture_name);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST + config.gl_filter_type);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST + config.gl_filter_type);
		glBindTexture(GL_TEXTURE_2D, 0);	// unbound
	}
}

void EMU_OSD::release_opengl()
{
	if (!use_opengl) {
		return;
	}
	release_texture();
}

void EMU_OSD::release_texture()
{
	if (mix_texture_name != 0 && glIsTexture(mix_texture_name) == GL_TRUE) {
		// release texture
		glDeleteTextures(1, &mix_texture_name);
	}
	mix_texture_name = 0;
}

void EMU_OSD::set_opengl_attr()
{
}

void EMU_OSD::set_opengl_poly(int width, int height)
{
}

#endif // USE_OPENGL
