/** @file wxw_screenmode.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2019.10.01

	@brief [ screen mode ]
*/

#include <wx/wx.h>
#include "../screenmode.h"
#include "../../emu.h"
#include <wx/display.h>
#include <wx/gdicmn.h>


ScreenMode::ScreenMode()
{
	disp_device_count = 0;
	monitor_count = 0;
	screen_mode_count = 0;
}

ScreenMode::~ScreenMode()
{
}

void ScreenMode::Enum(int desktop_width, int desktop_height, uint32_t flags)
{
	// enumerate display device
	int max_desktop_width = 0;
	int max_desktop_height = 0;

	// enumerate screen mode for fullscreen

	// the number of connected display monitors
	int disp_max_num = (int)wxDisplay::GetCount();
	if (disp_max_num > DISP_DEVICE_MAX) {
		disp_max_num = DISP_DEVICE_MAX;
	}
	for(int disp_no = 0; disp_no < disp_max_num; disp_no++) {
		wxDisplay disp(disp_no);
		wxVideoMode cmode = disp.GetCurrentMode();
		RECT_IN(disp_device[disp_no].re, 0, 0, cmode.GetWidth(), cmode.GetHeight());
	}

	screen_mode_count = 0;

	for(int disp_no = 0; disp_no < disp_max_num; disp_no++) {
		wxDisplay disp(disp_no);

		// get video modes
		wxArrayVideoModes vmodes = disp.GetModes();
		wxVideoMode cmode = disp.GetCurrentMode();

		logging->out_debugf(_T("screen_mode: current [%d] %dx%d %dbpp %dHz"), disp_no, cmode.GetWidth(), cmode.GetHeight(), cmode.GetDepth(), cmode.GetRefresh());

		for(int i = 0; i < (int)vmodes.Count() && screen_mode_count < SCREEN_MODE_MAX; i++) {
			int w = vmodes[i].GetWidth();
			int h = vmodes[i].GetHeight();
			int d = vmodes[i].GetDepth();
			int r = vmodes[i].GetRefresh();

			if(w < MIN_WINDOW_WIDTH || h < MIN_WINDOW_HEIGHT || w < 640 || h < 480) {
				logging->out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp %dHz ignored"), disp_no, w, h, d, r);
				continue;
			}
			if(((w * 30 / h) != 40)
			&& ((w * 40 / h) != 50)
			&& ((w * 90 / h) != 160)
			&& ((w * 50 / h) != 80)) {
				logging->out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp %dHz ignored"), disp_no, w, h, d, r);
				continue;
			}
			if (w <= 1280 && (w != 640 && w != 800 && w != 1024 && w != 1280)) {
				logging->out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp %dHz ignored"), disp_no, w, h, d, r);
				continue;
			}
			if (d != cmode.GetDepth() || r < cmode.GetRefresh() - 1 || cmode.GetRefresh() + 1 < r) {
				logging->out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp %dHz ignored"), disp_no, w, h, d, r);
				continue;
			}
			bool found = false;
			for(int j = 0; j < screen_mode_count; j++) {
				if(screen_mode[j].disp_no == disp_no && screen_mode[j].width == w && screen_mode[j].height == h) {
					found = true;
					logging->out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp %dHz already exist"), disp_no, w, h, d, r);
					break;
				}
			}
			if(!found) {
				screen_mode[screen_mode_count].disp_no = disp_no;
				screen_mode[screen_mode_count].width = w;
				screen_mode[screen_mode_count].height = h;
//				screen_mode_wx[screen_mode_count] = vmodes[i];
				logging->out_debugf(_T("screen_mode:%2d [%d] %dx%d %dbpp %dHz"), screen_mode_count, disp_no, w, h, d, r);
				screen_mode_count++;

				if(max_desktop_width <= w) {
					max_desktop_width = w;
				}
				if(max_desktop_height <= h) {
					max_desktop_height = h;
				}
			}
		}
	}
	qsort(&screen_mode, screen_mode_count, sizeof(struct screen_mode_st), qsort_by_n_w_h);

	// if cannot get modes, add default screen size mode.
	if (screen_mode_count == 0) {
		screen_mode[screen_mode_count].disp_no = 0;
		screen_mode[screen_mode_count].width = desktop_width;
		screen_mode[screen_mode_count].height = desktop_height;
		wxVideoMode cmode(desktop_width, desktop_height, wxDisplayDepth(), 60);
//		screen_mode_wx[screen_mode_count] = cmode;
		logging->out_debugf(_T("screen_mode:%2d [%d] %dx%d %dbpp %dHz"), screen_mode_count, 0, cmode.GetWidth(), cmode.GetHeight(), cmode.GetDepth(), cmode.GetRefresh());
		screen_mode_count = 1;

		max_desktop_width = desktop_width;
		max_desktop_height = desktop_height;
	}
}

int ScreenMode::qsort_by_n_w_h(const void *a, const void *b)
{
	struct screen_mode_st *a1 = (struct screen_mode_st *)a;
	struct screen_mode_st *b1 = (struct screen_mode_st *)b;

	int rc = 0;
	int aa = a1->width * a1->height;
	int bb = b1->width * b1->height;
	if (!rc) rc = (a1->disp_no - b1->disp_no);
	if (!rc) rc = (bb - aa);
	return rc;
}

const screen_mode_t *ScreenMode::GetMode(int num) const
{
	if (0 <= num && num < screen_mode_count) {
		return &screen_mode[num];
	}
	return NULL;
}

int ScreenMode::FindMode(int disp_no, int width, int height) const
{
	int match = -1;
	for(int i = 0; i < screen_mode_count; i++) {
		if (screen_mode[i].disp_no == disp_no
		&&  screen_mode[i].width == width
		&&  screen_mode[i].height == height) {
			match = i;
			break;
		}
	}
	return match;
}

const disp_device_t *ScreenMode::GetDisp(int num) const
{
	if (0 <= num && num < disp_device_count) {
		return &disp_device[num];
	}
	return NULL;
}
